import pandas as pd
import mappy
import numpy as np
import pysam
def rev_comp(strand):
    '''
    Get the reverse complement of a given sequence
    '''
    comp = ""
    for i in range(0, len(strand)):
        next_base = strand[len(strand) - i - 1]
        #check what the base is to switch to complement
        if next_base == "A":
            comp = comp + "T"
        if next_base == "T":
            comp = comp + "A"
        if next_base == "G":
            comp = comp + "C"
        if next_base == "C":
            comp = comp + "G"
        if next_base == "N":
            comp = comp + "N"
    return comp

def seq2int(seq):
  '''
  Convert DNA sequence to integer representation for Viterbi algorithm
  '''
  d = {'A': '1', 'T': '2', 'C': '3', 'G': '4'}
  seqInt = ' '.join(d[s] if s in d else s for s in seq.upper())
  return seqInt
  
def read_fasta(fileObject):
  '''
  Generator function to read in reads from fasta
  '''
  header = ''
  seq = ''
  # skip any useless leading information
  for line in fileObject:
    if line.startswith('>'):
      header = line.strip()
      break
  for line in fileObject:
    if line.startswith('>'):
      if 'N' in seq.upper():
        yield header, seq, False
      else:
        yield header, seq, True
      header = line.strip()
      seq = ''
    else:
      seq += line.strip()
  if header:
    if 'N' in seq.upper():
      yield header, seq, False
    else:
      yield header, seq, True

def read_fastq(fileObject):
  '''
  Generator function to read in reads from fastq
  '''
  header = ''
  seq = ''
  # skip any useless leading information
  for line in fileObject:
    #print(line)
    if line.startswith('@'):
      header = line.strip()
      break
  for line in fileObject:
    if line.startswith('@'):
      if 'N' in seq.upper():
        yield header, seq, False
      else:
        yield header, seq, True
      header = line.strip()
      seq = ''
    #check if optional or quality field
    elif line.startswith('+'):
       #assume quailty is one line and is after optional line
       next(fileObject)
       continue
    else:
      seq += line.strip()
  if header:
    if 'N' in seq.upper():
      yield header, seq, False
    else:
      yield header, seq, True

def remove_outlier_IQR(df,quantile=0.25):
    '''
    Removes outlier repeat copy numbers from distribution
    '''
    Q1=df.quantile(quantile)
    Q3=df.quantile(1-quantile)
    IQR=Q3-Q1
    df_final=df[~((df<(Q1-1.5*IQR)) | (df>(Q3+1.5*IQR)))]
    outliers= df[~(df.isin(df_final))]
    return df_final, outliers

def throw_low_cov(df):
    '''
    Filters any copy number with less than one supporting read
    '''
    filtered = df[df.freq > 1]
    return filtered

def get_IQR(df,quantile=0.25):
    '''
    Calculates the inter quantile range of the supporting read distribution
    '''
    Q1=df.quantile(quantile)
    Q3=df.quantile(1-quantile)
    IQR=Q3-Q1
    return (Q1, Q3), (Q1-1.5*IQR, Q3+1.5*IQR)

#input adapters for probabilities
def read_model_params(input_file, param_type):
   '''
   Converts user tsv input of custom model paramaters to dictionaries compatible with the
   original implementation
   Args:
      input_file (str): file name of input parameter tsv
      param_type (str): string designating which parameter dictionary to convert to, options: transitions, emissions, repeats, background
    Returns:
      param_dict (dictionary): dictionary of designated parameters
      
   '''
   #convert to respective dictionary
   if param_type in ["transitions","background"]: #input_df is a series here
    #convert to transition dictionary
    input_df = pd.read_csv(input_file, sep="\t")
    param_dict = input_df.iloc[0].to_dict()
    if param_type == 'background':
      param_dict[''] = 0
   if param_type in ["emissions","repeat"]: #input_df is a dataframe here
    #convert to transition dictionary
    input_df = pd.read_csv(input_file, sep="\t",index_col=0)
    if param_type == 'repeat':
       #add blank row
      input_df.loc[len( pd.DataFrame(input_df).index)] = 0
      input_df.index = ['A','G','C','T','']
    param_dict = input_df.to_dict()
   return param_dict

def generate_input_sheet(coords_file, chrom_sizes_file, ref,flanking_length=200):
  '''
  Function to generate input sheet with the following info from 4 or 5 column bedfile with the following columns: chromosome, start, end, repeat motif, name
  (name is an optional column with default name being the coordinated in chr:start-end format)
  Args:
    coords_file (str): bed file with 4 or 5 columns in chromosome, start, end, repeat motif, name format
    chrom_sizes_file (str): chrom.sizes file containing the sizes of each chromosome in the reference, used to ensure prefix/suffix does not exceed the length of the chromosome
    ref (str): path to reference genome to use
    flanking_length (int): length of desired prefix and suffix, must be greater than the length of flank parameter (if given) or 30 (if default), recommended >100bp
  Returns:
    targets (DataFrame): pandas dataframe of targets with columns name, prefix, repeat, suffix
  '''
  #load reference genome
  ref_aligner = mappy.Aligner(ref)
  #load chromosome sizes
  chrom_sizes = pd.read_csv(chrom_sizes_file, sep="\t",names=["chr","chr_length"])
  #load bedfile
  coords=pd.read_csv(coords_file,sep="\t",header=None)
  #check if name column is included
  if len(coords.columns) == 5:
    coords.columns =  ["chr","start","end","repeat","name"]
  elif len(coords.columns) == 4:
    #add name column
    coords.columns = ["chr","start","end","repeat"]
    coords["name"] = coords.chr + ":" +coords.start.astype(str) + "-" + coords.end.astype(str)
  else:
    print("Bed file has unexpected number of columns, please ensure file has 4 (repeat motif) or 5 (motif and name) columns. Exiting...")
    return
  #get prefix and suffix coordinates
  coords['prefix_start'] = coords.start - flanking_length
  coords['suffix_end'] = coords.end + flanking_length
  #replace invalid flanking starts and ends
  coords_final = pd.merge(left=coords,right=chrom_sizes,how="left",on="chr")
  coords_final['prefix_start'] = np.where(coords_final['prefix_start'] < 0,0,coords_final['prefix_start'])
  coords_final['suffix_end'] = np.where(coords_final['suffix_end'] > coords_final['chr_length'], coords_final['chr_length'], coords_final['suffix_end'])

  #set up dictionary
  targets_dict = {"name":[],"prefix":[],"repeat":[],"suffix":[]}
  for target in coords_final.itertuples():
    targets_dict['name'].append(target.name)
    targets_dict["prefix"].append(ref_aligner.seq(name=target.chr,start=target.prefix_start,end=target.start))
    targets_dict['repeat'].append(target.repeat)
    targets_dict['suffix'].append(ref_aligner.seq(name=target.chr,start=target.end,end=target.suffix_end))
  return pd.DataFrame(targets_dict)

def remove_flanking_outliers(counts_data):
  '''
  filter outlier reads based on their flanking sequence likelihood.
  this is useful for filtering off target reads and reads with large insertions
  in their repeat sequence
  '''
  counts_data["flanking_likelihood"] = counts_data["subset_likelihood"] - counts_data["repeat_likelihood"]
  thresh = counts_data["flanking_likelihood"].mean() - counts_data["flanking_likelihood"].std()
  df_final=counts_data[counts_data.flanking_likelihood > thresh]
  outliers= counts_data[~(counts_data.read_id.isin(df_final.read_id))]
  return df_final, outliers.read_id

def sort_alleles(curr_row, new_row, bootstrap, allele_specif_CIs): #this is only for diploid, need to translate to any allele number
    #establish an order of alleles based on median
    allele_med_cols = curr_row[curr_row.index.str.endswith("median")]
    allele_med_cols_order = allele_med_cols.sort_values(ascending=False) #largest to smallest, can change
    allele_med_cols_order = allele_med_cols.sort_values() 
    for i,allele in enumerate(allele_med_cols_order):
        #get current corresponding original row
        curr_allele = allele_med_cols_order.index[i].split(":")[0]
        curr_allele_cols = curr_row[curr_row.index.str.contains(curr_allele)]
        new_row["A"+str(i+1)+":median"] = curr_allele_cols[curr_allele + ":median"]
        new_row["A"+str(i+1)+":mode"] = curr_allele_cols[curr_allele + ":mode"]
        new_row["A"+str(i+1)+":SD"] = curr_allele_cols[curr_allele + ":SD"]
        new_row["A"+str(i+1)+":supporting_reads"] = curr_allele_cols[curr_allele + ":supporting_reads"]
        if bootstrap:
          new_row["A"+str(i+1)+":median_CI"] = curr_allele_cols[curr_allele + ":median_CI"] 
        if allele_specif_CIs:
          curr_CI = curr_allele_cols[curr_allele + ":median_CI_allele_specific"][0] #tuple, take lower bound and find closest allele assignment
          closest_allele_bootstrap = allele_med_cols_order[(allele_med_cols_order-curr_CI).abs().argsort()].index[0].split(":")[0]
          new_row["A"+ str(i+1)+":median_CI_allele_specific"] = curr_row[closest_allele_bootstrap + ":median_CI_allele_specific"]
    return new_row

def sort_outputs(curr_row, max_peaks,bootstrap,allele_specif_CIs, stranded=False):
  '''
  Use this to sort the column names to be in the same order and order alleles from smallest to largest 
  '''
  column_names_in_order = []
  column_names_in_order.append("name")
  if stranded:
      column_names_in_order.append("strand")
  for i in range(1,max_peaks+1):
      column_names_in_order.append("A"+str(i)+":median")
      column_names_in_order.append("A"+str(i)+":mode")
      column_names_in_order.append("A"+str(i)+":SD")
      column_names_in_order.append("A"+str(i)+":supporting_reads")
      if bootstrap:
        column_names_in_order.append("A"+str(i)+":median_CI")
      if allele_specif_CIs:
        column_names_in_order.append("A"+ str(i)+":median_CI_allele_specific")
  column_names_in_order.append("num_supporting_reads")
  column_names_in_order.append("bandwidth")
  column_names_in_order.append("peak_calling_method")
  new_row = curr_row[column_names_in_order].copy()
  new_row_final = sort_alleles(curr_row, new_row,bootstrap, allele_specif_CIs)
  return new_row_final

def sort_read_assign_alleles(read_assign, geno, max_k):
    '''
    Sorts the alleles in the read_assign file to match the geno file first.
    This function now takes max_k as an argument to dynamically handle the number of alleles.
    '''
    # Generate the list of allele median columns dynamically based on max_k
    allele_columns = [f"A{i}:median" for i in range(1, max_k + 1)]
    
    # Calculate medians for read_assign
    read_assign_allele_medians = read_assign.groupby(["name", "cluster_assignments"]).counts.median().reset_index()
    
    # Melt the geno dataframe to long format for easier comparison, using the dynamically generated allele columns
    geno_allele_medians = geno.melt(id_vars=["name"], value_vars=allele_columns, var_name="allele", value_name="median")
    geno_allele_medians["allele"] = geno_allele_medians["allele"].str.extract('A(\d+):median').astype(int)
    
    # Copy read_assign to avoid modifying the original dataframe
    new_read_assign = read_assign.copy()
    
    # Iterate over unique names
    unique_names = read_assign["name"].unique()
    for name in unique_names:
        geno_allele_medians_subset = geno_allele_medians[geno_allele_medians["name"] == name]
        for allele in read_assign["cluster_assignments"].dropna().unique():
            read_median = read_assign[(read_assign["name"] == name) & (read_assign["cluster_assignments"] == allele)].counts.median()
            closest_allele = geno_allele_medians_subset.iloc[(geno_allele_medians_subset["median"]-read_median).abs().argsort()[:1]]["allele"].to_list()[0]
            new_read_assign["cluster_assignments"] = np.where((read_assign["name"] == name) & (read_assign["cluster_assignments"] == allele), closest_allele, new_read_assign["cluster_assignments"])
    
    return new_read_assign

def get_read_ids(bam_file, chrom,start,end,region_name):
    # Open the BAM file
    with pysam.AlignmentFile(bam_file, "rb") as bam:
        # Get all reads that align to the specified region
        reads = bam.fetch(chrom, start,end)
        # Get the IDs of the reads
        read_ids = {read.query_name:region_name for read in reads}
    return read_ids
def bam_assign(bam_file,target_coordinates):
    '''
    This function assigns reads to targets based on their alignment to the reference genome
    this mode will be optional and fo now its an additional rather than an alternative input
    will be used to pre-assign reads from global alignment to targets to avoid misassignemnt during WGS runs
    or runs with a large number of targets with non-unique direct flanking sequence
    This mode will also only allow for one-to-one assignment while local alignment only allows multiple
    which is suboptimal for a loarge number of targets in this iteration of the software
    '''
    #for every target region, get all reads that align to it and add them to a dictionary of read_ids to target names
    targets = pd.read_csv(target_coordinates,sep="\t", header=None)
    #check if there is a 5th column in the bed file
    if len(targets.columns) == 5:
        targets.columns = ["chr","start","end","repeat","name"]
    else:
        targets.columns = ["chr","start","end","repeat"]
        targets["name"] = targets.chr + ":" +targets.start.astype(str) + "-" + targets.end.astype(str)
    #for every target in targets, get the reads that align to it
    read_ids = {}
    for target in targets.itertuples():
        read_ids.update(get_read_ids(bam_file, target.chr ,target.start, target.end,target.name))
    return read_ids




import seaborn as sns
from sklearn.mixture import GaussianMixture
from scipy.stats import norm
import scipy.stats as st
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import pickle as pkl
import math
from matplotlib.ticker import MaxNLocator
import warnings
warnings.filterwarnings('ignore')

from HMMSTR_utils.HMMSTR_utils import remove_outlier_IQR, throw_low_cov

class GMMStats:
    def __init__(self, target_row):
        self.name = target_row[0]
        self.prefix = target_row[1].rstrip().upper()[-30:]
        self.repeat = target_row[2].rstrip().upper()
        self.suffix = target_row[3].rstrip().upper()[:30]
    
    def plot_stats(self,counts_data,out):
        '''
        plot the histograms and median likelihood ratio line plots (hopefully with error bars)

        '''
        #(1) unmodified histogram
        to_plot = counts_data[['counts','freq']].drop_duplicates().sort_values(by="counts")
        if to_plot.shape[0] == 0 or to_plot.shape[1] == 0:
            return
        #(1) unmodified histogram
        sns.reset_orig()
        to_plot.plot.bar(x="counts",y="freq",figsize=(15,5),title= str(self.name), color="red", fontsize=7)
        plt.xlabel("Repeat count")
        plt.ylabel("Number of reads")
        fig = plt.gcf()
        fig.savefig(str(out +self.name+ "_supporting_reads_hist.pdf"))
        plt.show()
        plt.clf()
        return #currently does not work and isnt called
    def get_stats(self, out_count_file, out, plot_hists, filter_outliers=False):
        '''
        Function to calculate summary stats for a single target

        Args:
            out_count_name(str): output suffix for final result tsv

        Returns:
            counts_data(pandas dataframe): final calculations for given count file

        '''
        #read output file
        counts_data = pd.read_csv(out_count_file, sep=" ",header=None)
        counts_data.columns = ["read_id","strand","align_score","neg_log_likelihood","subset_likelihood","repeat_likelihood","align_start", "align_end","counts"]
        counts_data['freq'] = counts_data.groupby(by='counts')['read_id'].transform('count')
        #ADDED outlier filtering
        outliers = []
        if filter_outliers:
            filtered, outliers = remove_outlier_IQR(counts_data.counts)
            #counts_data = counts_data[counts_data.counts.isin(filtered)]
        counts_data['outlier'] = np.where(counts_data.counts.isin(outliers),True , False)
        #testing out including plotting here, will need to add a flag
        if plot_hists:
            self.plot_stats(counts_data, out)
        return counts_data

    def call_peaks(self,X_orig,out, num_max_peaks, plot=True, save_allele_plots = True):
        # https://www.cbrinton.net/ECE20875-2020-Spring/W11/gmms_notebook.pdf
        rng = np.random.RandomState(seed=1)

        #get transformed counts column to represent weighted frequencies
        X = X_orig.copy()
        X = X[["counts"]]
        if X.shape[0] < 2:
            curr_dict = {"name":self.name}
            #revised
            #check if we have 1 read
            if X.shape[0] == 1:
                # curr_dict["H1:mean"] = X.counts[0]
                curr_dict["H1:median"] = X.counts[0]
                curr_dict["H1:mode"] = X.counts[0]
                curr_dict["H1:SD"] = 0
                curr_dict["H1:supporting_reads"] = 1
                cluster_assignments = np.array([0])
            else:
                cluster_assignments = np.array([-1])
            for i in range(1,num_max_peaks+1):
                #curr_mean = "H"+str(i)+":mean"
                curr_median = "H"+str(i)+":median"
                curr_mode = "H"+str(i)+":mode"
                curr_sd = "H" + str(i) + ":SD"
                curr_support = "H" + str(i) + ":supporting_reads"
                #if curr_mean not in curr_dict.keys():
                   # curr_dict[curr_mean] = 0 #don't include the mean since our distributions are left skewed, report median and mode
                if curr_median not in curr_dict.keys():
                    curr_dict[curr_median] = 0
                if curr_mode not in curr_dict.keys():
                    curr_dict[curr_mode] = 0
                if curr_sd not in curr_dict.keys():
                    curr_dict[curr_sd] = 0
                if curr_support not in curr_dict.keys():
                    curr_dict[curr_support] = 0
        # for i in range(1,num_max_peaks+1):
        #     curr_dict["H"+str(i)+":SD"] = final_peaks_SD[i-1]
            curr_dict["num_supporting_reads"] = len(X_orig.index)
            return pd.Series(curr_dict), cluster_assignments
        # Fit models with 1-10 components
        #account for more components than samples
        if X.shape[0] < num_max_peaks: # FIXME this may be an issue in consistency, maybe save original number for output
            num_max_peaks = X.shape[0]

        k_arr = np.arange(num_max_peaks) + 1
        models = [
        GaussianMixture(n_components=k).fit(X)
        for k in k_arr
        ]
        dfs = [] #subsets of X corresponding to each peak
        # Plot function
        def plot_mixture(gmm, X, show_legend=True, ax=None, title=None, subplot=False,cluster=None):
            if plot:
                plt.rcParams["figure.figsize"] = (10,6)
                sns.set_style('whitegrid',{'grid.color': '#dcddde'}) # nice and clean grid
                if ax is None:
                    ax = plt.gca()

            # Compute PDF of whole mixture
            x = np.linspace(X.min(), X.max(), 1000)
            logprob = gmm.score_samples(x.reshape(-1, 1))
            pdf = np.exp(logprob)

            # Compute PDF for each component
            responsibilities = gmm.predict_proba(x.reshape(-1, 1))

            pdf_individual = responsibilities * pdf[:, np.newaxis]
            #print(pdf_individual.shape)
            #get the max of each pdf
            maxInColumns = np.amax(pdf_individual, axis=0)

            # Plot data histogram
            if plot:
                ax.xaxis.set_major_locator(MaxNLocator(integer=True))
                hist = ax.hist(X.counts, density=True, histtype='bar',  label='Data', bins=range(int(min(X.counts)),int(max(X.counts))+2),facecolor = '#9bc1cc', edgecolor='whitesmoke', linewidth=0.5, align="left") #updated plotting parameters
                ax.set_xlabel("Repeat Copy Number")
                ax.set_ylabel("Density")
                # Plot PDF of whole model
                if not subplot:
                    ax.plot(x, pdf, '-k', label='Mixture PDF',linewidth=2)

                # Plot PDF of each component, divided so we can call this multiple times, once for full mixture and once for each cluster
                if subplot:
                    ax.plot(x, pdf_individual[:,cluster],color="black")
                else:
                    ax.plot(x, pdf_individual, '--', label='Component PDF')
                if title is not None:
                    ax.set_title(title)
                if subplot:
                    #add metrics to plot
                    ymin, ymax = ax.get_ylim()
                    ax.set_ylim(ymin, ymax)
                    ax.vlines(x = st.mode(np.array(X.counts), keepdims=True)[0][0],ymin=ymin, ymax=ymax,linewidth=2, color='r',linestyle="dashed", label="Mode")
                    #ax.axvline(x = X.counts.mean(), linewidth=2, color='g', linestyle="dashed", label = "Mean")
                    ax.axvline(x = X.counts.median(), linewidth=2, color='purple',linestyle="dashed",label="Median")
                    #print("mean", X.counts.mean(), " mode: ", st.mode(np.array(X.counts))[0][0], "median: ", X.counts.median())
                if show_legend:
                    ax.legend()
                if not subplot: #save full peaks plot
                    plt.savefig(out + self.name + "peaks.pdf",bbox_inches='tight')
                    plt.show()
                    plt.clf()
                else:
                    plt.savefig(out + self.name + "allele_"+ str(i+1)+".pdf",bbox_inches='tight')
                    plt.show()
                    plt.clf()
                plt.show()
                plt.clf()
            return gmm.means_[np.argsort(maxInColumns)],np.argsort(maxInColumns)
        AIC = [m.aic(X) for m in models]
        BIC = [m.bic(X) for m in models]

        # Plot these metrics
        if plot:
            fig = plt.gcf()
            fig.set_size_inches(15,7)
            plt.plot(k_arr, AIC, label='AIC')
            plt.plot(k_arr, BIC, label='BIC')
            plt.xlabel('Number of Components ($k$)')
            plt.legend()
            plt.tight_layout()
            plt.title(self.name+": AIC and BIC")
            plt.savefig(out + self.name + "AIC_BIC.pdf",bbox_inches='tight')
            plt.show()
            plt.clf()
            fig = plt.gcf()
            fig.set_size_inches(15,7)
        gmm_best = models[np.argmin(AIC)]
        max_peaks, ranks_idx = plot_mixture(gmm_best, X, title=self.name +": Gaussian Mixture Model of Best Fit")
        #print(max_peaks)
        # if plot:
        #     plt.title(self.name)
        #     plt.savefig(out + self.name + "peaks.pdf",bbox_inches='tight')
        #     plt.show()
        #     plt.clf()

        #ADDED 9/7/22 TO TEST CLUSTERING + NEW FITTING WITH SCIPY
        cluster_assignments = np.argmax(gmm_best.predict_proba(X), axis=1)
        for cluster in np.unique(cluster_assignments):
            dfs.append(X[np.where(cluster_assignments == cluster, True, False)])
        #plot separate clusters
        curr_dict = {"name":self.name}
        cluster_stats = {}
        for i,cluster in enumerate(dfs):
            #plot distributions separately
            cluster_stats[i] = {}
            if save_allele_plots and plot:
                test1, test2 = plot_mixture(gmm_best, cluster,title=self.name+": Model of Best Fit for Allele " + str(i+1), subplot=True,cluster=i)
                #plt.title(self.name)
                #plt.savefig(out + self.name + "allele_"+ str(i+1)+".pdf",bbox_inches='tight')
                #plt.show()
                #plt.clf()
            #add metrics to dictionary so they are outputted
            #cluster_stats[i]['mean'] = cluster.counts.mean()
            cluster_stats[i]['median'] = cluster.counts.median()
            cluster_stats[i]['mode'] = st.mode(np.array(cluster),keepdims=True)[0][0][0]
            if math.isnan(cluster.counts.std()):
                cluster_stats[i]['SD']  = 0
            else:
                cluster_stats[i]['SD'] = cluster.counts.std()
            cluster_stats[i]['supporting_reads'] = len(np.where(cluster_assignments == i)[0])
        #unravel cluster_stats dictionary for curr_row
        for i in cluster_stats.keys():
            #curr_dict["H" + str(i+1) + ":mean"] = cluster_stats[i]["mean"]
            curr_dict["H" + str(i+1) + ":median"] = cluster_stats[i]["median"]
            curr_dict["H" + str(i+1) + ":mode"] = cluster_stats[i]["mode"]
            curr_dict["H" + str(i+1) + ":SD"] = cluster_stats[i]["SD"]
            curr_dict["H" + str(i+1) + ":supporting_reads"] = cluster_stats[i]["supporting_reads"]

        #check for missing columns and fill them if not there
        for i in range(1,num_max_peaks+1):
            #curr_mean = "H"+str(i)+":mean"
            curr_median = "H"+str(i)+":median"
            curr_mode = "H"+str(i)+":mode"
            curr_sd = "H" + str(i) + ":SD"
            curr_support = "H" + str(i) + ":supporting_reads"
            #if curr_mean not in curr_dict.keys():
                #curr_dict[curr_mean] = 0
            if curr_median not in curr_dict.keys():
                curr_dict[curr_median] = 0
            if curr_mode not in curr_dict.keys():
                curr_dict[curr_mode] = 0
            if curr_sd not in curr_dict.keys():
                curr_dict[curr_sd] = 0
            if curr_support not in curr_dict.keys():
                curr_dict[curr_support] = 0
        curr_dict["num_supporting_reads"] = len(X_orig.index)
            
        curr_row = pd.Series(curr_dict)
        # print(curr_row, "length: ", len(curr_row))
        # print(cluster_assignments)
        return curr_row, cluster_assignments
    
    def bootstrap_gmm(self,curr_row,data, resample_size, CI_width, max_peaks, out):
        #FIXME need to determine if I should only construct CIs from each allele separately instead of like how I am since if there is a large imbalance in supporting reads, the CIs will include both alleles by chance
        #throw out 1 read coverage, this is something to assume will happen if bootstrapping
        filtered = throw_low_cov(data)
        call_lists = {}
        for i in range(1,max_peaks+1):
            #call_lists["H"+str(i)+"_mean"]=[]
            call_lists["H"+str(i)+"_median"]=[]
            #call_lists["H"+str(i)+"_mode"]=[]

        for i in range(resample_size):
            #sample up to the depth of the experiment with replacement
            curr_sample = filtered.sample(frac=1,axis=0, replace = True)
            #call gmm with original freqs on this subset
            curr_row = self.call_peaks(curr_sample, out ,max_peaks, plot=False) 
            #add current call to lists
            for j in range(1,max_peaks+1):
                #call_lists["H"+str(j)+"_mean"].append(curr_row[0]['H'+str(j)+":mean"])
                call_lists["H"+str(j)+"_median"].append(curr_row[0]['H'+str(j)+":median"])
                #call_lists["H"+str(j)+"_mode"].append(curr_row[0]['H'+str(j)+":mode"])
        #mean_CIs = {}
        median_CIs = {}
        #mode_CIs = {}
        adjustment = float((1.0 - CI_width)/2)
        for j in range(1,max_peaks+1): #get confidence interval for each allele called
            #curr_means = call_lists["H"+str(j)+"_mean"]
            curr_medians = call_lists["H"+str(j)+"_median"]
            #curr_modes = call_lists["H"+str(j)+"_mode"]
            #curr_row[0]["H"+str(j)+":mean_CI"] = (np.quantile(curr_means,adjustment),np.quantile(curr_means,1-adjustment))
            curr_row[0]["H"+str(j)+":median_CI"] = (np.quantile(curr_medians,adjustment),np.quantile(curr_medians,1-adjustment))
            #curr_row[0]["H"+str(j)+":mode_CI"] = (np.quantile(curr_modes,adjustment),np.quantile(curr_modes,1-adjustment))
        #print(curr_row)
        return curr_row

    def bootstrap_gmm_allele_specific(self,data, resample_size, CI_width, out):
            #FIXME need to determine if I should only construct CIs from each allele separately instead of like how I am since if there is a large imbalance in supporting reads, the CIs will include both alleles by chance
            #throw out 1 read coverage, this is something to assume will happen if bootstrapping
            filtered = throw_low_cov(data)
            call_lists = {}
            #call_lists["mean"]=[]
            call_lists["median"]=[]
            #call_lists["mode"]=[]
    
            for i in range(resample_size):
                #sample up to the depth of the experiment with replacement
                curr_sample = filtered.sample(frac=1,axis=0, replace = True)
                #call gmm with original freqs on this subset
                curr_row = self.call_peaks(curr_sample, out ,1, plot=False) #single allele bootstrapping
                #add current call to lists
                #call_lists["mean"].append(curr_row[0]["H1:mean"])
                call_lists["median"].append(curr_row[0]["H1:median"])
                #call_lists["mode"].append(curr_row[0]["H1:mode"])
            #mean_CIs = {}
            median_CIs = {}
            #mode_CIs = {}
            adjustment = float((1.0 - CI_width)/2)
            #curr_means = call_lists["mean"]
            curr_medians = call_lists["median"]
            #curr_modes = call_lists["mode"]
            #mean_CI = (np.quantile(curr_means,adjustment),np.quantile(curr_means,1-adjustment))
            median_CI= (np.quantile(curr_medians,adjustment),np.quantile(curr_medians,1-adjustment))
            #mode_CI = (np.quantile(curr_modes,adjustment),np.quantile(curr_modes,1-adjustment))
            return median_CI


from setuptools import setup, Extension
#from Cython.Build import cythonize
import subprocess
import os

#Custom function to find install location of glib-2.0
def pkgconfig(*packages):
    cmd = ['pkg-config', '--cflags', '--libs'] + list(packages)
    output = subprocess.check_output(cmd).decode().split()
    cflags = [flag[2:] for flag in output if flag.startswith('-I')]
    libs = [flag[2:] for flag in output if flag.startswith('-l')]
    return cflags, libs

#added [glib_cflags[0]+"/glib"] to designate where the included/hiddedn header files are for glib-2.0, TODO make sure this works for the install I include in README
glib_cflags, glib_libs = pkgconfig('glib-2.0')
try:
    from Cython.Build import cythonize
    extensions = cythonize([
    Extension(
        "c_files.hmm",
        ["src/c_files/hmm.pyx","src/c_files/sequence.c","src/c_files/viterbi.c","src/c_files/hmmutils.c","src/c_files/nrutil.c","src/c_files/hmmrand.c"],
        libraries=glib_libs,
        include_dirs=["src/c_files"]+glib_cflags+[glib_cflags[0]+"/glib"],
        extra_compile_args=[],
        extra_link_args=[],
    ),
])
except ImportError:
    extensions = [
    Extension(
        "c_files.hmm",
        ["src/c_files/hmm.c","src/c_files/sequence.c","src/c_files/viterbi.c","src/c_files/hmmutils.c","src/c_files/nrutil.c","src/c_files/hmmrand.c"],
        libraries=glib_libs,
        include_dirs=["src/c_files"]+glib_cflags+[glib_cflags[0]+"/glib"],
        extra_compile_args=[],
        extra_link_args=[],
    ),
]


setup(
    name="HMMSTR",
    version="1.0.4",
    python_requires='>=3.8.17',
    packages=["HMMSTR","GMM_stats","HMMSTR_utils","c_files","process_vit_res","profile_HMM","process_read","KDE_stats"],
    ext_modules=extensions,
    install_requires=['colorama','numpy','pandas','pickleshare','scikit-learn','scipy','seaborn','importlib-resources','mappy','pysam'],
    zip_safe=False,
    scripts=['src/c_files/test_hmm_cython.py'],
    # package_data = {
    #     'c_files': ['*.pxd']},
)


# Import the converted pxd declarations
cimport c_files.hmm as chmm
import cython
from libc.stdio cimport FILE,stdout, sscanf, fclose, fopen, fprintf, stderr
from libc.stdlib cimport malloc, free, atoi
# Wrapper functions for the HMM functions
cdef void py_ReadHMM(FILE* fp, chmm.HMM* phmm):
    chmm.ReadHMM(fp, phmm)

cdef void py_PrintHMM(FILE* fp, chmm.HMM* phmm):
    chmm.PrintHMM(fp, phmm)

cdef void py_InitHMM(chmm.HMM* phmm, int N, int M, int seed):
    chmm.InitHMM(phmm, N, M, seed)

cdef void py_CopyHMM(chmm.HMM* phmm1, chmm.HMM* phmm2):
    chmm.CopyHMM(phmm1, phmm2)

cdef void py_FreeHMM(chmm.HMM* phmm):
    chmm.FreeHMM(phmm)

cdef void py_PrintDelta(FILE* fp, double** delta, chmm.HMM* phmm, int T):
    chmm.PrintDelta(fp, delta, phmm, T)

cdef void py_PrintTraceBack(FILE* fp, int** delta, chmm.HMM* phmm, int T):
    chmm.PrintTraceBack(fp, delta, phmm, T)

cdef void py_ReadSequence(FILE* fp, int* pT, int** pO):
    chmm.ReadSequence(fp, pT, pO)

cdef void py_ReadSequences(FILE* fp, int* pN, int** pT, int*** pO):
    chmm.ReadSequences(fp, pN, pT, pO)

cdef void py_PrintSequences(int* pN, int* size, int*** O):
    chmm.PrintSequences(pN, size, O)

cdef void py_PrintSequence(FILE* fp, int T, int* O):
    chmm.PrintSequence(fp, T, O)

cdef void py_PrintPath(FILE* fp, int T_new, chmm.GList* list):
    chmm.PrintPath(fp, T_new, list)

cdef void py_PrintLikelihoods(FILE* fp, int T_new, chmm.GList* likelihoods):
    chmm.PrintLikelihoods(fp, T_new, likelihoods)

cdef void py_GenSequenceArray(chmm.HMM* phmm, int seed, int T, int* O, int* q):
    chmm.GenSequenceArray(phmm, seed, T, O, q)

cdef int py_GenInitalState(chmm.HMM* phmm):
    return chmm.GenInitalState(phmm)

cdef int py_GenNextState(chmm.HMM* phmm, int q_t):
    return chmm.GenNextState(phmm, q_t)

cdef int py_GenSymbol(chmm.HMM* phmm, int q_t):
    return chmm.GenSymbol(phmm, q_t)

cdef void py_PrintAllOuts(FILE* fp, int T_new, chmm.GList* list, double** delta, int N, int T):
    chmm.PrintAllOuts(fp, T_new, list, delta, N, T)

cdef void py_Viterbi(chmm.HMM* phmm, int T, int* O, double** delta, int** psi, int** traceback_dir, chmm.GList** list, int* T_new, int* q, double* pprob):
    chmm.Viterbi(phmm, T, O, delta, psi, traceback_dir, list, T_new, q, pprob)

cdef void py_ViterbiLog(chmm.HMM* phmm, int T, int* O, double** delta, int** psi, int** traceback_dir, chmm.GList** list, chmm.GList** likelihoods_ints, chmm.GList** likelihoods_dec, int* T_new, int* q, double* pprob, int repeat_len, int prefix_idx):
    chmm.ViterbiLog(phmm, T, O, delta, psi, traceback_dir, list, likelihoods_ints, likelihoods_dec, T_new, q, pprob, repeat_len, prefix_idx)

cdef int py_hmmgetseed():
    return chmm.hmmgetseed()

cdef void py_hmmsetseed(int seed):
    chmm.hmmsetseed(seed)

cdef double py_hmmgetrand():
    return chmm.hmmgetrand()

cdef void py_free(void* ptr):
    free(ptr)
cdef void cy_main(bytes model_file, bytes obs_seq_length, bytes obs_seq,bytes repeat_len_str,bytes prefix_idx_str):
    cdef int t, T, T_new, repeat_len, prefix_idx
    cdef chmm.HMM hmm
    cdef int* O
    cdef int* q
    cdef double** delta
    cdef int** psi
    cdef int** traceback_dir
    cdef double proba, logproba
    cdef FILE *fp, *traceback_del, *traceback, *delta_out
    cdef int i, n, offset
    cdef char* S
    cdef chmm.GList *list_log = NULL
    cdef chmm.GList *likelihoods_ints = NULL
    cdef chmm.GList *likelihoods_dec = NULL

    #if argc != 6:
     #   print("Usage error")
      #  print("Usage: testvit <model.hmm> <obs.seq length> <obs.seq>")
      #  return

    # open HMM file
    fp = fopen(model_file, "r")
    if fp == NULL:
        fprintf(stderr, "Error: File %s not found\n", model_file)
        return

    # read HMM file according to ReadHMM
    chmm.ReadHMM(fp, &hmm)
    fclose(fp)

    T = atoi(obs_seq_length) # read T from input
    S = obs_seq
    repeat_len =  atoi(repeat_len_str)# for calculation of RDn case
    prefix_idx = atoi(prefix_idx_str)# for calculation of RDn case

    O = chmm.ivector(1, T)
    i = 1
    while (sscanf(S, "%d%n", &n, &offset) == 1):
        O[i] = n
        S += offset
        i += 1

    q = chmm.ivector(1, T)
    traceback_dir = chmm.imatrix(1, T, 1, hmm.N)
    delta = chmm.dmatrix(1, T, 1, hmm.N)
    psi = chmm.imatrix(1, T, 1, hmm.N)

    chmm.ViterbiLog(&hmm, T, O, delta, psi, traceback_dir, &list_log,&likelihoods_ints,&likelihoods_dec, &T_new, q, &logproba, repeat_len, prefix_idx)

    fprintf(stdout, "Viterbi  MLE log prob = %E\n", logproba)
    fprintf(stdout, "Optimal state sequence:\n")
    chmm.PrintPath(stdout, T_new, list_log)
    chmm.PrintLikelihoods(stdout, T_new, likelihoods_ints)
    chmm.PrintLikelihoods(stdout, T_new, likelihoods_dec)

    chmm.free_ivector(q, 1, T)
    chmm.free_ivector(O, 1, T)
    chmm.free_imatrix(psi, 1, T, 1, hmm.N)
    chmm.free_imatrix(traceback_dir, 1, T, 1, hmm.N)
    chmm.g_list_free(list_log);
    chmm.g_list_free(likelihoods_ints);
    chmm.g_list_free(likelihoods_dec);
    chmm.free_dmatrix(delta, 1, T, 1, hmm.N)
    chmm.py_FreeHMM(&hmm)

# Example usage of the HMM struct and functions
def example_usage():
    cdef chmm.HMM hmm
    cdef int T = 10
    cdef int* O_ptr = <int*>malloc((T+1) * sizeof(int))
    cdef int* p_ptr = <int*>malloc((T+1) * sizeof(int))
    # Initialize the HMM
    chmm.py_InitHMM(&hmm, 3, 2, 123)
    print("made it past initHMM")
    # Generate a sequence using the HMM
    chmm.py_GenSequenceArray(&hmm, 456, T, O_ptr, p_ptr)
    print("made it past py_GenSequenceArray")
    # Convert the sequence to a Python list
    O = [O_ptr[i] for i in range(1,T+1)]
    p = [p_ptr[i] for i in range(1,T+1)]
    print(p)
    print("made it past making O a list")
    # Print the generated sequence
    print("Generated Sequence:")
    print(O)

    # Clean up
    chmm.py_FreeHMM(&hmm)
    free(O_ptr)
    free(p_ptr)
    print("O freed")

# Run the example usage
#example_usage()
def run_main(model_file,obs_seq_length,obs_seq,repeat_len,prefix_idx):
    # ...
    model_file_bytes = model_file.encode('utf-8')
    obs_seq_length_bytes = obs_seq_length.encode('utf-8')
    obs_seq_bytes = obs_seq.encode('utf-8')
    repeat_len_bytes = repeat_len.encode('utf-8')
    prefix_idx_bytes = prefix_idx.encode('utf-8')

    cy_main(model_file_bytes, obs_seq_length_bytes, obs_seq_bytes, repeat_len_bytes, prefix_idx_bytes)

    #model_file = b'test_out/CANVAS_random_coord/test_outs_CANVAS.hmm'
    #obs_seq_length = b'831'
    #obs_seq = b'2 3 4 2 1 1 1 1 1 3 4 4 1 1 4 1 4 1 2 4 1 1 3 2 1 1 2 4 1 1 4 1 1 1 3 3 4 4 2 3 1 4 3 3 1 1 1 1 4 4 4 1 3 3 1 4 1 1 3 2 2 4 1 1 4 1 2 2 2 4 2 1 1 1 1 2 2 3 2 3 1 4 3 3 2 4 2 3 3 4 2 1 2 2 2 3 1 4 1 1 1 1 2 4 1 1 1 1 1 4 3 1 1 4 2 2 3 2 1 4 4 3 3 1 4 4 2 4 2 4 4 2 4 4 3 2 3 1 3 4 3 3 2 4 2 1 1 2 3 3 3 1 4 3 1 2 2 4 2 4 4 4 1 4 1 3 4 4 4 3 3 1 1 2 3 1 3 2 2 3 1 4 4 2 3 1 4 4 1 4 2 2 2 4 1 4 1 3 3 1 4 3 2 2 4 4 3 3 1 1 2 1 2 4 4 2 1 1 1 1 1 3 3 3 3 1 2 3 2 3 2 1 4 4 1 1 1 1 1 2 1 3 1 1 1 1 1 2 2 1 4 3 2 4 4 1 3 1 2 4 4 2 4 4 3 1 3 4 3 4 1 3 2 1 2 1 4 2 3 3 3 1 4 3 2 1 3 2 3 1 4 4 1 4 4 3 2 4 1 4 4 3 1 4 4 1 4 1 2 2 3 1 3 2 2 4 1 1 3 3 2 4 4 4 1 4 4 3 1 4 1 4 4 2 2 4 3 1 4 2 4 1 4 3 3 4 1 4 1 2 2 4 3 4 3 3 1 3 2 4 3 1 3 2 3 3 1 4 2 3 2 4 4 4 3 1 1 3 1 4 1 4 3 1 1 4 1 3 2 3 2 4 2 2 2 3 1 1 1 1 1 1 1 1 4 1 1 1 1 4 1 1 1 1 1 4 1 1 1 1 4 1 1 1 1 4 1 1 1 1 4 1 1 1 4 1 1 1 1 1 4 3 1 2 4 2 2 3 2 1 1 1 4 1 4 1 1 2 1 4 3 1 1 3 4 4 2 4 3 1 4 3 2 4 4 1 4 2 1 2 3 1 2 4 1 2 1 4 1 4 1 2 2 1 2 4 4 4 1 2 2 1 2 1 3 1 4 4 3 1 1 1 1 1 3 1 3 2 4 2 3 1 4 2 2 2 4 1 1 2 2 4 1 1 4 1 4 4 1 2 4 4 1 4 1 1 1 4 4 1 3 1 1 1 1 2 4 1 1 1 4 1 1 4 4 3 2 4 2 3 1 4 1 3 2 2 3 2 4 1 4 1 2 2 3 3 1 4 2 4 1 3 1 4 4 1 3 1 1 3 1 3 1 4 3 2 1 2 2 3 2 2 3 1 1 4 1 1 1 1 1 4 4 4 1 1 4 1 1 2 4 4 3 3 3 3 1 1 1 1 4 3 1 1 2 2 3 1 4 1 4 1 2 3 3 1 1 3 3 1 2 3 1 4 1 1 3 2 3 3 3 2 3 2 2 3 1 1 1 1 4 1 2 4 4 1 4 3 2 2 4 4 2 2 2 3 1 1 3 1 4 1 2 4 1 4 1 3 1 4 3 3 1 3 3 1 3 3 3 3 1 1 1 4 3 3 2 2 4 2 4 4 4 3 1 4 4 1 2 4 4 3 3 3 1 4 3 1 4 1 4 3 2 4 3 1 4 1 1 4 3 1 4 4 1 3 1 4 3 2 4 3 3 2 4 2 4 1 3 2 4 2 4 4 4 4 4 2 1 4 4 4 3 3 1 3 2 1 1 3 3 3 1 4 2 4 4 4 2 3 2 1 4 1 4 4 4 3 1 4 4 4 3 1 2 3 1 1 4 3 3'
    #repeat_len = b'5'
    #prefix_idx = b'91'

    #cy_main(model_file, obs_seq_length, obs_seq,repeat_len,prefix_idx)

    # ...



import seaborn as sns
import scipy.stats as st
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import warnings
from numpy import array, linspace
from sklearn.neighbors import KernelDensity
from scipy.signal import argrelextrema
import os
import statistics as stat
from collections import Counter

from HMMSTR_utils.HMMSTR_utils import throw_low_cov, remove_outlier_IQR, remove_flanking_outliers
warnings.filterwarnings('ignore')

class KDE_cluster:
    def __init__(self, target, out, out_count_name, discard_outliers,flanking_like_filter,curr_strand=None):
        self.name = target[0]
        self.out = out
        self.discard_outliers = discard_outliers #this will be used when we goto cluster
        self.flanking_like_filter = flanking_like_filter
        out_count_file = out + "_" + self.name + out_count_name
        if os.path.exists(out_count_file):
            counts_data = pd.read_csv(out_count_file, sep=" ",header=None)
            counts_data.columns = ["read_id","strand","align_score","neg_log_likelihood","subset_likelihood","repeat_likelihood","repeat_start","repeat_end","align_start", "align_end","counts"]
            self.data = counts_data[counts_data.counts != 0] #discard 0 count reads
            if curr_strand is not None:
                self.strand = curr_strand
                self.data = counts_data[counts_data.counts != 0][counts_data.strand == curr_strand] #if this is a stranded run, we only want to keep data from the current strand
            else:
                self.strand = None
            if self.data.empty:
                self.data = None
        else:
            self.data = None
    def get_stats(self, plot_hists, filter_outliers=False, filter_quantile=0.25,flanking_like_filter=False, strand=None):
        '''
        Function to calculate summary stats for a single target

        Args:
        -----------------------------------------------------------------------------------------------
            plot_hists (bool): boolean indicating if read support histograms should be saved
            filter_outliers (bool): boolean indicating if outliers should be included in allele calls
            filter_quantile (float): if filter_outliers is set, this quantile of repeat copy number will be filtered
            flanking_like_filter (bool): if set, reads with excessively low likelihood in their flanking regions will be filtered
            strand (str): which strand to operated on if applicable

        Returns:
        -----------------------------------------------------------------------------------------------
            counts_data(pandas dataframe): final calculations for given count file

        '''
        #read output file
        counts_data = self.data
        counts_data['freq'] = counts_data.groupby(by='counts')['read_id'].transform('count')
        #outlier filtering
        outliers = []
        flanking_outliers=[]
        if filter_outliers:
            filtered, outliers = remove_outlier_IQR(counts_data.counts, filter_quantile)
            #counts_data = counts_data[counts_data.counts.isin(filtered)]
        counts_data['outlier'] = np.where(counts_data.counts.isin(outliers),True , False)
        #check for flanking likelihood outliers
        if flanking_like_filter:
            flanking_filtered, flanking_outliers = remove_flanking_outliers(counts_data)
        counts_data['flanking_outlier'] = np.where(counts_data.read_id.isin(flanking_outliers),True , False)
        if plot_hists:
            self.plot_stats(counts_data, self.out, strand)
        return counts_data
    def plot_stats(self,counts_data,out, strand=None):
        '''
        plot the histograms

        Args:
        -----------------------------------------------------------------------------
        counts_data (dataframe): dataframe continaing copy numbers from Viterbi output
        out (str): output prefix
        strand (str): strand we are plotting if applicable

        Returns:
        ----------------------------------------------------------------------------
        None

        '''
        to_plot = counts_data[['counts','freq']].drop_duplicates().sort_values(by="counts")
        if to_plot.shape[0] == 0 or to_plot.shape[1] == 0:
            return
        sns.reset_orig()
        to_plot.plot.bar(x="counts",y="freq",figsize=(15,5),title= str(self.name), color="red", fontsize=7)
        plt.xlabel("Repeat count")
        plt.ylabel("Number of reads")
        fig = plt.gcf()
        if self.strand is not None:
            fig.savefig(str(self.out + "_plots/"+self.name+ "_"+ self.strand +"_supporting_reads_hist.pdf"))
        else:
            fig.savefig(str(self.out + "_plots/"+self.name+"_supporting_reads_hist.pdf"))
        plt.show()
        plt.clf()
        return #currently does not work and isnt called
    def call_clusters(self,kernel="gaussian", bandwidth = 'scott', max_k = 2, output_plots = False, subset=None, filter_quantile=0.25, allele_specific = False, allele = None, strand=None):
        '''
        This function uses kernal density estimation to resolve alleles. It is ideal when data is normally distributed
        and has a relatively small range or overlapping count distributions

        Args
        --------------------------------------------------------------------------------------------------------------
        kernel: str
            String designating kernel to use for the kde
        bandwidth: float or str
            Bandwidth to use in kde
        max_k: int
            The maximum number of clusters to keep in outputs, all others are filtered by cluster size
        output_plots: bool
            True if density plot will be saved, False if not
        subset: bool
            if True, cluster a subset of the data
        filter_quantile: float
            qunatile to filter if filter outliers is set
        allele_specific: bool
            if True, plot allele specific plots
        allele: str
            allele to plot and cluster
        strand: str
            which strand to plot and cluster if applicable

        Returns
        -------------------------------------------------------------------------------------------------------------
        clusters: dict
            Dictionary containing all counts assigned to each cluster
        allele_calls: dict
            Dictionary containing the allele call from each cluster as well as the size and standard deviation within the cluster
        outliers: pandas Series
            Series of counts that are identified as outliers and were dicarded before clustering. Empty if filter_outliers is False
        flanking_outliers: pandas Series
            Series of reads to filter based on the likelihood of the viterbi path through the prefix and suffix

        '''
        #filter outliers if called for
        outliers = pd.Series()
        flanking_outliers = pd.Series()
        if self.discard_outliers and self.flanking_like_filter == False: #only discard outliers
            filtered, outliers = remove_outlier_IQR(self.data.counts,filter_quantile)
            a = np.array(filtered).reshape(-1,1)
        elif self.flanking_like_filter: #either flanking filter or both
            flanking_filtered, flanking_outliers = remove_flanking_outliers(self.data)
            #check if we also filtered outliers so we can do both
            if self.discard_outliers:
                filtered, outliers = remove_outlier_IQR(flanking_filtered.counts,filter_quantile) #filter outliers
                a = np.array(flanking_filtered[flanking_filtered.counts.isin(filtered.unique())].counts).reshape(-1,1) #use reads that are both in both of our filtered sets
            else:
                a = np.array(flanking_filtered.counts).reshape(-1,1)

        else:
            a = np.array(self.data.counts).reshape(-1,1)
        if subset is not None:
            a = np.array(subset.counts).reshape(-1,1) #allow a subset of data to be passed for reclustering
        try:#initial call
            kde = KernelDensity( bandwidth=bandwidth,kernel=kernel).fit(a)

            #trying minimum bandwidth
            if kde.bandwidth_ < 0.5:
                kde = KernelDensity( bandwidth=0.5,kernel=kernel).fit(a)
        except:
            if len(a) == 0:
                clusters = -1
                allele_calls = {"A"+str(k)+":median":-1 for k in range(1,max_k+1)}
                allele_calls.update({"A"+str(k)+":mode":-1 for k in range(1,max_k+1)})
                allele_calls.update({"A"+str(k)+":SD":-1 for k in range(1,max_k+1)})
                allele_calls.update({"A"+str(k)+":supporting_reads":-1 for k in range(1,max_k+1)})
                allele_calls["bandwidth"] = -1
                return clusters,allele_calls , outliers, flanking_outliers

            print("Switching to a constant bandwidth=0.5")
            kde = KernelDensity( bandwidth=0.5,kernel=kernel).fit(a)
        s = linspace(min(a)-2,max(a)+2)
        e = kde.score_samples(s.reshape(-1,1))
        mi, ma = argrelextrema(e, np.less)[0], argrelextrema(e, np.greater)[0]

        #run and change bandwidth till # maxima matches or is less than max_k
        while len(mi)+1 > max_k:
            bandwidth = kde.bandwidth_ + 0.1
            kde = KernelDensity( bandwidth=bandwidth,kernel=kernel).fit(a)
            e = kde.score_samples(s.reshape(-1,1))
            mi, ma = argrelextrema(e, np.less)[0], argrelextrema(e, np.greater)[0]

        #save cluster info based on minima and maxima
        all_clusters = {}
        maxima = {}
        sizes = {}
        if len(mi) == 0:
            all_clusters["1"] = a
            maxima["1"] = s[ma]
            sizes["1"] = len(a)
        elif len(mi) == 1:
            #print("hi")
            all_clusters["1"] = a[a < s[mi][0]]
            maxima["1"] = s[ma][0]
            sizes["1"] = len(a[a < s[mi][0]])
            all_clusters["2"] = a[a >= s[mi][0]]
            if len(s[ma]) > 1:
                maxima["2"] = s[ma][1]
            else:
                if isinstance(stat.mode(all_clusters["2"]), np.float64):
                    maxima["2"] = stat.mode(all_clusters["2"])
                else:
                    maxima["2"] = stat.mode(all_clusters["2"])[0] #if multiple modes returned, pick the first
            sizes["2"] = len(a[a >= s[mi][0]])
        else: # more than 2 clusters
            all_clusters["1"] = a[a < s[mi][0]]
            maxima["1"] = s[ma][0]
            sizes["1"] = len(a[a < s[mi][0]])
            for i_cluster in range(len(mi)-1):
                all_clusters[str(i_cluster+2)] = a[(a >= s[mi][i_cluster]) * (a <= s[mi][i_cluster+1])]
                maxima[str(i_cluster+2)] = s[ma][i_cluster+1]
                sizes[str(i_cluster+2)] = len(a[(a >= s[mi][i_cluster]) * (a <= s[mi][i_cluster+1])])
            all_clusters[str(len(mi)+1)] = a[a >= s[mi][-1]]
            maxima[str(len(mi)+1)] = s[ma][-1]
            sizes[str(len(mi)+1)] = len(a[a >= s[mi][-1]])

        #filter clusters based on max_k
        sizes_series = pd.Series(sizes)
        #sort sizes
        sizes_series_sorted = sizes_series.sort_values(ascending=False)
        #get max_k clusters by size
        k_clusters = list(sizes_series_sorted.head(max_k).index)
        clusters = {k: np.unique(all_clusters[k]) for k in k_clusters}
        non_one_maxima = False
        for k in k_clusters:
            if isinstance(maxima[k], np.float64) == False and len(maxima[k]) != 1:
                non_one_maxima = True
        if non_one_maxima == False:
            allele_calls = {}
            allele_calls.update({"A" +str(k)+":supporting_reads":sizes[k] for k in k_clusters})
            allele_calls.update({"A" +str(k)+":SD":np.std(all_clusters[k]) for k in k_clusters})
            allele_calls.update({"A" +str(k)+":median":np.median(all_clusters[k]) for k in k_clusters})
            allele_calls.update({"A" +str(k)+":mode":int(st.mode(np.array(all_clusters[k]),keepdims=True)[0][0]) for k in k_clusters}) 
            
            allele_calls["bandwidth"] = kde.bandwidth_
        elif len(k_clusters) == 1:
            #this is the case when there is only 1 value detected thus it is the only allele that can be called
            allele_calls = {}
            allele_calls.update({"A1:supporting_reads":sizes["1"]})
            allele_calls.update({"A1:SD":np.std(all_clusters["1"])})
            allele_calls.update({"A1:median":np.median(all_clusters["1"])})
            allele_calls.update({"A1:mode":int(st.mode(np.array(all_clusters["1"]),keepdims=True)[0][0])})
            allele_calls["bandwidth"] = kde.bandwidth_
        else:
            allele_calls = {}
            allele_calls["bandwidth"] = -1
            allele_calls.update({"A"+str(k)+":SD":0 for k in k_clusters})
            allele_calls.update({"A"+str(k)+":median":0 for k in k_clusters})
            allele_calls.update({"A"+str(k)+":mode":0 for k in k_clusters})
            return clusters,allele_calls , outliers, flanking_outliers

        if output_plots:
            if len(mi) == 0:
                try:
                    plt.hist(a, density=True, histtype='bar',  label='Data', bins=range(int(min(a)),int(max(a))+2),facecolor = '#9bc1cc', edgecolor='whitesmoke', linewidth=0.5, align="left")
                    plt.plot(s[:],np.exp(e[:]),color="black",lw=2,linestyle="-",label="Kernel Density Estimate")
                        
                    if len(s[ma] != 0):
                        plt.plot(s[ma], np.exp(e[ma]), 'bo',label="KDE Maxima")
                    
                except:
                    print("Encounted an exception in homozygous plotting case (KDE)")
                if allele_specific:
                    plt.title(self.name + " KDE: Allele " + allele)
                    ymin, ymax = plt.ylim()
                    plt.ylim(ymin, ymax)
                    plt.vlines(x = allele_calls["A1:mode"],ymin=ymin, ymax=ymax,linewidth=2, color='r',linestyle="dashed", label="Mode")
                    plt.vlines(x = allele_calls["A1:median"], ymin=ymin, ymax=ymax,linewidth=2, color='purple',linestyle="dashed",label="Median")
                    plt.legend()  
                    fig = plt.gcf()
                    if strand is not None:
                        fig.savefig(str(self.out + "_plots/"+self.name +"_" +strand+"_allele_"+ allele+"_KDE.pdf"))
                    else:
                        fig.savefig(str(self.out + "_plots/"+self.name + "_allele_"+ allele+"_KDE.pdf"))
                else:
                    plt.title(self.name + " KDE")
                    plt.legend()  
                    fig = plt.gcf()
                    if strand is not None:
                        fig.savefig(str(self.out + "_plots/"+self.name +"_" +strand+ "_KDE.pdf"))
                    else:
                        fig.savefig(str(self.out + "_plots/"+self.name + "_KDE.pdf"))
                plt.show()
                plt.clf()
            elif len(mi) == 1:
                try:
                    plt.hist(a, density=True, histtype='bar',  label='Data', bins=range(int(min(a)),int(max(a))+2),facecolor = '#9bc1cc', edgecolor='whitesmoke', linewidth=0.5, align="left")
                    plt.plot(s[:],np.exp(e[:]),color="black",lw=2,linestyle="-",label="Kernel Density Estimate")
                    plt.plot(s[ma], np.exp(e[ma]), 'bo',label="KDE Maxima")
                except:
                    print("An exception has occured")
                plt.plot(s[mi], np.exp(e[mi]), 'ro',label="KDE Minima")
                if allele_specific:
                    plt.title(self.name + " KDE: Allele "+ allele)
                    ymin, ymax = plt.ylim()
                    plt.ylim(ymin, ymax)
                    plt.vlines(x = allele_calls["A1:mode"],ymin=ymin, ymax=ymax,linewidth=2, color='r',linestyle="dashed", label="Mode")
                    plt.vlines(x = allele_calls["A1:median"], ymin=ymin, ymax=ymax,linewidth=2, color='purple',linestyle="dashed",label="Median")
                    plt.legend()  
                    fig = plt.gcf()
                    if strand is not None:
                        fig.savefig(str(self.out + "_plots/"+ self.name +"_"+ strand+"_allele_"+ allele+"_KDE.pdf"))
                    else:
                        fig.savefig(str(self.out + "_plots/"+ self.name + "_allele_"+ allele+"_KDE.pdf"))

                else:
                    plt.title(self.name + " KDE")
                    plt.legend()  
                    fig = plt.gcf()
                    if strand is not None:
                        fig.savefig(str(self.out + "_plots/"+ self.name +"_"+ strand+ "_KDE.pdf"))
                    else:
                        fig.savefig(str(self.out + "_plots/"+ self.name + "_KDE.pdf"))
                plt.show()
                plt.clf()
            #dynamically print clusters
            else:
                try:
                    plt.hist(a, density=True, histtype='bar',  label='Data', bins=range(int(min(a)),int(max(a))+2),facecolor = '#9bc1cc', edgecolor='whitesmoke', linewidth=0.5, align="left")
                    plt.plot(s[:],np.exp(e[:]),color="black",lw=2,linestyle="-",label="Kernel Density Estimate")
                    plt.plot(s[ma], np.exp(e[ma]), 'bo',label="KDE Maxima")
                except:
                    print("An exception occured in plotting the KDE")
                plt.plot(s[mi], np.exp(e[mi]), 'ro',label="KDE Minima")
                if allele_specific:
                    plt.title(self.name + " KDE: Allele " +allele)
                    ymin, ymax = plt.ylim()
                    plt.ylim(ymin, ymax)
                    plt.vlines(x = allele_calls["A1:mode"],ymin=ymin, ymax=ymax,linewidth=2, color='r',linestyle="dashed", label="Mode")
                    plt.vlines(x = allele_calls["A1:median"], ymin=ymin, ymax=ymax,linewidth=2, color='purple',linestyle="dashed",label="Median")
                    plt.legend()  
                    fig = plt.gcf()
                    if strand is not None:
                        fig.savefig(str(self.out + "_plots/"+self.name + "_"+ strand+"_allele_"+ allele+"_KDE.pdf"))
                    else:
                        fig.savefig(str(self.out + "_plots/"+self.name + "_allele_"+ allele+"_KDE.pdf"))

                else:
                    plt.title(self.name + " KDE")
                    plt.legend()  
                    fig = plt.gcf()
                    if strand is not None:
                        fig.savefig(str(self.out + "_plots/"+ self.name + "_"+ strand+"_KDE.pdf"))
                    else:
                        fig.savefig(str(self.out + "_plots/"+ self.name + "_KDE.pdf"))
                plt.show()
                plt.clf()
        return clusters, allele_calls, outliers, flanking_outliers
    def assign_clusters(self, clusters, outliers, flanking_outliers):
        '''
        Function to assign individual reads to clusters determined by KDE

        Parameters
        ------------------------------------------------------------------
        clusters: dict
            Dictionary of counts assigned to each cluster
        outliers: series
            series containing all outliers identified
        flanking_outliers: series
            series containing all flanking sequence likelihood outliers
        
        Returns:
        ----------------------------------------------------------------------
        assignemnt_df: pandas DataFrame
            DataFrame equivalent to input count data with additional columns for cluster
            assignemnts and if the count was an outlier
        '''
        count_assignemnts = {}
        if clusters == -1:
            return pd.DataFrame(columns =['read_id', 'strand', 'align_score', 'neg_log_likelihood', 'subset_likelihood', 'repeat_likelihood', 'repeat_start', 'repeat_end', 'align_start', 'align_end', 'counts', 'freq', 'cluster_assignments', 'outlier', 'flanking_outlier'])
        for cluster in clusters:
            for curr in clusters[cluster]:
                count_assignemnts[curr] = cluster
        labels = pd.DataFrame()
        labels['counts'] = count_assignemnts.keys()
        labels['cluster_assignments'] = count_assignemnts.values()
        assignment_df = self.data.merge(labels, how = 'left', on = 'counts')
        assignment_df['outlier'] = np.where(assignment_df.counts.isin(outliers),True , False)
        assignment_df['flanking_outlier'] = np.where(assignment_df.read_id.isin(flanking_outliers),True , False)
        return assignment_df

    def bootstrap_KDE(self,data, resample_size, CI_width, max_peaks, out):
        '''
        Performs bootstrapping and constructs confidence intervals using GMM peakcalling

        Args:
        --------------------------------------------------------------------------------
        data (dataframe): counts dataframe to sample from
        resample_size (int): number of times to resample during bootstraping
        CI_width (float): width of confidience interval to calculate
        max_peaks (int): maximum number of peaks to call in each bootstrap iteration
        out (str): output prefix

        Returns:
        --------------------------------------------------------------------------------
        curr_row (series): the original input row with bootstrap results included
        '''
        filtered = throw_low_cov(data)
        call_lists = {}
        for i in range(1,max_peaks+1):
            call_lists["A"+str(i)+":median"]=[]

        for i in range(resample_size):
            if resample_size > 100 and (i+1)%10 == 0:
                print("Round " + str(i+1)+ "/" + str(resample_size))
            elif resample_size < 100 and resample_size > 10 and (i+1)%5 == 0:
                print("Round " + str(i+1)+ "/" + str(resample_size))
            elif resample_size < 10:
                print("Round " + str(i+1)+ "/" + str(resample_size))
            #sample up to the depth of the experiment with replacement
            curr_sample = filtered.sample(frac=1,axis=0, replace = True)
            #call gmm with original freqs on this subset
            clusters, allele_calls, outliers, flanking_outliers = self.call_clusters(kernel="gaussian", bandwidth = 'scott', max_k = max_peaks, output_plots = False, subset=curr_sample, filter_quantile=0.25) 
            #add current call to lists
            for j in range(1,max_peaks+1):
                if ("A"+str(j)+":median") not in allele_calls.keys():
                    call_lists["A"+str(j)+":median"].append(0)
                    allele_calls["A"+str(j)+":median"] = 0
                    continue
                call_lists["A"+str(j)+":median"].append(allele_calls["A"+str(j)+":median"])
        median_CIs = {}
        adjustment = float((1.0 - CI_width)/2)
        for j in range(1,max_peaks+1): #get confidence interval for each allele called
            curr_medians = call_lists["A"+str(j)+":median"]
            median_CIs["A"+str(j)+":median_CI"] = (np.quantile(curr_medians,adjustment),np.quantile(curr_medians,1-adjustment))
        return median_CIs

    def bootstrap_KDE_allele_specific(self,data, resample_size, CI_width, out):
            #throw out 1 read coverage, this is something to assume will happen if bootstrapping
            filtered = throw_low_cov(data)
            call_lists = {}
            call_lists["median"]=[]
    
            for i in range(resample_size):
                if resample_size > 100 and (i+1)%10 == 0:
                    print("Round " + str(i+1)+ "/" + str(resample_size))
                elif resample_size < 100 and resample_size > 10 and (i+1)%5 == 0:
                    print("Round " + str(i+1)+ "/" + str(resample_size))
                elif resample_size < 10:
                    print("Round " + str(i+1)+ "/" + str(resample_size))
                #sample up to the depth of the experiment with replacement
                curr_sample = filtered.sample(frac=1,axis=0, replace = True)
                #call gmm with original freqs on this subset
                clusters, allele_calls, outliers, flanking_outliers = self.call_clusters(kernel="gaussian", bandwidth = 'scott', max_k = 1, output_plots = False, subset=curr_sample, filter_quantile=0.25)  #single allele bootstrapping
                #add current call to lists
                call_lists["median"].append(allele_calls["A1:median"])
            adjustment = float((1.0 - CI_width)/2)
            curr_medians = call_lists["median"]
            median_CI= (np.quantile(curr_medians,adjustment),np.quantile(curr_medians,1-adjustment))
            return median_CI